#include <Python.h>
#include <methodobject.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "naw_api.h"
#include "defs.h"


/***************************************************************************
  Module methods, to be called from python by other files in the naw
  package.
 ***************************************************************************/

static PyObject *
verify (PyObject *self, PyObject *args)
{
    PyObject *err_type;
    int fd, type;

    if (!PyArg_ParseTuple(args, "Oii", &err_type, &fd, &type))
        return NULL;

    if (naw_verify(fd, type, 0) != fd) {
        PyErr_SetFromErrno(err_type);
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
verify_ex (PyObject *self, PyObject *args)
{
    PyObject *err_type;
    int fd, type;

    if (!PyArg_ParseTuple(args, "Oii", &err_type, &fd, &type))
        return NULL;

    if (naw_verify(fd, type, 0) != fd) {
        return PyInt_FromLong((long)errno);
    }

    return PyInt_FromLong(0L);
}


static char doc_verify[] =
    "Verify a currently connected socket.  On error, the descriptor\n"
    "will be closed and an exception, with errno carryover from\n"
    "naw_verify(), will be raised.";
static char doc_verify_ex[] =
    "Verify a currently connected socket.  On error, the descriptor\n"
    "will be closed and errno, as carried over from naw_verify(),\n"
    "will be returned.";

static PyMethodDef nawmod_methods[] = {
    {"verify_connection", verify, METH_VARARGS, doc_verify},
    {"verify_connection_ex", verify_ex, METH_VARARGS, doc_verify_ex},
    {}
};



/***************************************************************************
  Configuration functions and data.

  ConfSet maps strings to configuration options defined in naw_api.h.

  Functions beginning with '_' are used only by other nawConf functions.
  Functions without a leading underscore should be used as methods or
  mappings.
 ***************************************************************************/


#define NAW_TYPE_INT    0
#define NAW_TYPE_STRING 1

static struct nawConfArgs {
    char *name;
    int type;
    int vec_set;
    int vec_get;
} ConfSet[] = {
    { "CONFIG_PATH_CONNECT", NAW_TYPE_STRING,
                             NAW_SET_CONFIG_PATH_CONNECT,
                             NAW_GET_CONFIG_PATH_CONNECT },
    { "CONFIG_PATH_ACCEPT",  NAW_TYPE_STRING,
                             NAW_SET_CONFIG_PATH_ACCEPT,
                             NAW_GET_CONFIG_PATH_ACCEPT },
    { "FLUSH_TRIGGER_PATH",  NAW_TYPE_STRING,
                             NAW_SET_FLUSH_TRIGGER_PATH,
                             NAW_GET_FLUSH_TRIGGER_PATH },
    { "DEBUG_LEVEL",         NAW_TYPE_INT,
                             NAW_SET_DEBUG_LEVEL,
                             NAW_GET_DEBUG_LEVEL },
    { "DEBUG_NOISY",         NAW_TYPE_INT,
                             NAW_SET_DEBUG_NOISY,
                             NAW_GET_DEBUG_NOISY },
    {},
};

#define NAW_CONF_GET 0
#define NAW_CONF_SET 1


/***************************************************************************
  _nawConf_findset()

  Find a ConfSet matching the string [name].

  Returns a pointer to the associated ConfSet if found;
  0 otherwise.
 ***************************************************************************/

static inline struct nawConfArgs *
_nawConf_findset(const char *name)
{
    struct nawConfArgs *argp;

    for (argp = ConfSet; argp->name; argp++) {
        if (!strcmp(name, argp->name))
            return argp;
    }

    return 0;
}

/***************************************************************************
  _nawConf_runConf()
  _nawConf_runConf_internal()

  The body functions handling get and set operations of nawConfType

  Returns an object if successful (None for NAW_CONF_SET, or the result for
  NAW_CONF_GET); otherwise, returns 0.
 ***************************************************************************/

static PyObject *
_nawConf_runConf_internal(const char *argName, PyObject *argValue, int getset)
{
    struct nawConfArgs *opt;
    void  *confArg;

    if (!(opt = _nawConf_findset(argName))) {
        PyErr_SetString(PyExc_TypeError, "no such configuration option");
        return 0;
    }

    if (getset == NAW_CONF_SET) {
        if (!argValue) {
            PyErr_SetString(PyExc_TypeError, "unset not supported");
            return 0;
        }

        switch (opt->type) {
            case NAW_TYPE_STRING:
                confArg = PyString_AsString(argValue);
                if (!confArg) return 0;
                confArg = strdup(confArg); /* XXX: leak! do not cycle. */
                if (!confArg) return 0;
                /* if e above fails, we have bigger problems than an
                 * unspecified exception */
                break;
            case NAW_TYPE_INT:
                if (!PyInt_Check(argValue)) {
                    PyErr_SetString(PyExc_TypeError, "expecting integer value");
                    return 0;
                }
                confArg = (void *)(int)PyInt_AsLong(argValue);
                break;
        }
        naw_config(opt->vec_set, confArg);

        Py_INCREF(Py_None);
        argValue = Py_None;

    } else {
        if (argValue) {
            PyErr_SetString(PyExc_TypeError, "get method called with a value");
            return 0;
        }

        naw_config(opt->vec_get, &confArg);

        switch (opt->type) {
            case NAW_TYPE_STRING:
                argValue = PyString_FromString(confArg);
                break;
            case NAW_TYPE_INT:
                argValue = PyInt_FromLong((long) confArg);
                break;
        }
    }

    return argValue;
}


static inline PyObject *
_nawConf_runConf(PyObject *args, int getset)
{
    const char *argName;
    PyObject *argValue = 0;

    if (!PyArg_ParseTuple(args, "s|O", &argName, &argValue))
        return 0;

    return _nawConf_runConf_internal(argName, argValue, getset);
}

static PyObject *
nawConf_setconf(PyObject *self, PyObject *args)
{
    return _nawConf_runConf(args, NAW_CONF_SET);
}

static PyObject *
nawConf_getconf(PyObject *self, PyObject *args)
{
    return _nawConf_runConf(args, NAW_CONF_GET);
}

/***************************************************************************
  _nawConf_len_calc()

  Count how many configuration options are present.

  Returns the resultant long integer.
 ***************************************************************************/

static inline long
_nawConf_len_calc(void)
{
    struct nawConfArgs *argp = ConfSet;
    long len = 0;

    while (argp->name) len++, argp++;
    return len;
}

static PyObject *
nawConf_len(PyObject *self, PyObject *args)
{

    return PyInt_FromLong(_nawConf_len_calc());
}

static PyObject *
nawConf_contains(PyObject *self, PyObject *args)
{
    const char *argName;

    if (!PyArg_ParseTuple(args, "s", &argName))
        return NULL;

    if (!_nawConf_findset(argName)) {
        Py_INCREF(Py_False);
        return Py_False;
    }

    Py_INCREF(Py_True);
    return Py_True;
}

static PyObject *
nawConf_keys(PyObject *self, PyObject *args)
{
    PyObject *x, *l = PyList_New(_nawConf_len_calc());

    struct nawConfArgs *argp = ConfSet;
    int which = 0;

    while (argp->name) {
        x = PyString_FromString(argp->name);
        PyList_SetItem(l, which, x);
        which++, argp++;
    }

    return l;
}

static PyObject *
nawConf_notsupp(PyObject *self, PyObject *args)
{
    PyErr_SetString(PyExc_TypeError, "method not supported");
    return NULL;
}

static PyMethodDef nawConfMethods[] =
{
    { "set",          nawConf_setconf,  METH_VARARGS, 0 },
    { "get",          nawConf_getconf,  METH_VARARGS, 0 },
    { "__delitem__",  nawConf_notsupp,  METH_VARARGS, 0 }, /* XXX: useless */
    { "__len__",      nawConf_len,      METH_VARARGS, 0 },
    { "__contains__", nawConf_contains, METH_VARARGS, 0 }, /* XXX: useless */
    { "keys",         nawConf_keys,     METH_VARARGS, 0 },
    {},
};


/***************************************************************************
  Mappings for nawConfType, defined below.
 ***************************************************************************/

static int
nawConf_mp_len(PyObject *self)
{
    return (int) _nawConf_len_calc();
}

static PyObject *
nawConf_mp_get(PyObject *self, PyObject *key)
{
    char *ks = PyString_AsString(key);
    if (!ks) return NULL;

    return _nawConf_runConf_internal(ks, 0, NAW_CONF_GET);
}

static int
nawConf_mp_set(PyObject *self, PyObject *key, PyObject *value)
{
    char *ks;
    PyObject *rv;

    ks = PyString_AsString(key);
    if (!ks) return -1;

    rv = _nawConf_runConf_internal(ks, value, NAW_CONF_SET);
    if (!rv) return -1;

    Py_DECREF(rv);
    return 0;
}

static PyMappingMethods nawConf_as_mapping =
{
    nawConf_mp_len,  /* mp_length */
    nawConf_mp_get,  /* mp_subscript */
    nawConf_mp_set,  /* mp_ass_sub */
};


/***************************************************************************
  nawConfType

  Type for configuraation option (naw.conf) object.

  See NOTES for usage information.  Setting and querying configuration
  options is handled by dictionary-style subscripting.  Only valid
  configuration options may be subscripted.
 ***************************************************************************/

typedef struct {
    PyObject_HEAD
} nawConfObject;

static PyTypeObject nawConfType = {
    PyObject_HEAD_INIT(0)
    0,                     /* ob_size */
    "naw.nawconf",         /* tp_name */
    sizeof(nawConfObject), /* tp_basicsize */
};

PyMODINIT_FUNC
initnawinternal (void)
{
    PyObject *module;

    nawConfType.tp_flags      = Py_TPFLAGS_DEFAULT;
    nawConfType.tp_doc        = "naw type";
    nawConfType.tp_new        = PyType_GenericNew;
    nawConfType.tp_methods    = nawConfMethods;
    nawConfType.tp_as_mapping = &nawConf_as_mapping;

    if (PyType_Ready(&nawConfType) < 0)
        return;


    module = Py_InitModule("nawinternal", nawmod_methods);
    if (!module) return;

    PyModule_AddObject(module, "INIT_CLI", PyInt_FromLong((long)INIT_CLI));
    PyModule_AddObject(module, "INIT_SRV", PyInt_FromLong((long)INIT_SRV));
    PyModule_AddObject(module, "CONFIGDIR", PyString_FromString(NAW_CONFIGDIR));

    Py_INCREF(&nawConfType);
    PyModule_AddObject(module, "conf", (PyObject *)&nawConfType);

}

