#define SYSLOG_NAMES 1
#include <stdio.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "misc.h"
#include "lxmisc.h"
#include <lx_map.h>

#include "naw.h"
#include "naw_log.h"
#include "syslog.h"

/***************************************************************************
  naw_log_syslog_noop

  The no-op function.
 ***************************************************************************/

static int naw_log_syslog_noop (handle, data)
    struct naw_handle *handle;
    struct naw_data *data;
{
    return NAW_YES;
}

static void *oldlogptr;
static char *prefix;
static int facility = LOG_AUTH;
static ub1_t logdebuglevel = 3;

/***************************************************************************
  log a message to syslog

  NAW priority messages are translated to LOG priority messages, with the
  exception that NAW_DEBUG, NAW_DEBUG2, and NAW_DEBUG3 are all translated to
  LOG_DEBUG, and the logdebuglevel variable determines whether any of them
  are actually logged (see naw_log_syslog_init()).

  The 'prefix' variable determines the prefix for the log line (not the
  syslog tag, since the application may have used openlog(3) already), and
  'facility' determines the syslog facility.  See naw_log_syslog_init() for
  the defaults.

  If writing to syslog increases the file descriptor count (i.e., if the
  application has not yet used syslog, and this module's doing so opens
  another descriptor), closelog(3) is used to keep another descriptor from
  lingering where it may not be expected.
 ***************************************************************************/

static int naw_log_syslog(level, s)
    int level;
    lx_s *s;
{
    int debuglevel = level - NAW_DEBUG;
    int pfd[3];

    if (debuglevel++ < 0) debuglevel = 0;

    switch (level) {
        case NAW_CRIT:    level = LOG_CRIT;    break;
        case NAW_ERROR:   level = LOG_ERR;     break;
        case NAW_WARNING: level = LOG_WARNING; break;
        case NAW_NOTICE:  level = LOG_NOTICE;  break;
        case NAW_INFO:    level = LOG_INFO;    break;
        default:          level = LOG_DEBUG;   break;
    }

    if (lx_check0(s)) return NAW_ERR;

    log_debug (debuglevel, "log-syslog: ", s->s, 0,0,0);
    if (debuglevel > logdebuglevel) return NAW_OK;

    if (pipe(pfd)) {
        log_debug(1, "pipe creation failed",0,0,0,0);
        return NAW_ERR;
    }
    close(pfd[1]);
    close(pfd[0]);

    syslog(level|facility, "%s: %s", prefix, s->s);

    if (pipe(pfd+1)) {
        log_debug(1, "pipe creation failed",0,0,0,0);
        return NAW_ERR;
    }
    close(pfd[2]);
    close(pfd[1]);

    if (pfd[1] > pfd[0]) closelog();
    return NAW_OK;
}

/***************************************************************************
  initialize syslog logger

  arguments:
    log_prefix: name used as a prefix to each message (not the syslog tag)
                default: "libnaw"
    debuglevel: level of debugging to use as LOG_DEBUG
                0 means no debugging will be logged
                1 is NAW_DEBUG
                2 is NAW_DEBUG2
                3 is NAW_DEBUG3
                default: 3
    log_facility:
                syslog facility to use.  Arguments are strings as defined in
                'facilitynames' under sys/syslog.h.  If the facility named
                is unknown, the default will be used.
                default: "auth"
 ***************************************************************************/

static int naw_log_syslog_init (data)
    struct naw_data *data;
{
    char *v;
    int vlen;

    if (!naw_data_get(data, "log_prefix", &v, &vlen))
         prefix = strdup(v);
    else prefix = strdup("libnaw");
    if (!prefix) return NAW_ERR;

    if (!naw_data_get(data, "debuglevel", &v, &vlen))
        logdebuglevel = (ub1_t)strtoul(v, 0, 10);

    if (!naw_data_get(data, "log_facility", &v, &vlen)) {
        NAW_CODE *facilities = fac_names;
        for (;;) {
            if (!facilities->c_name) {
                NAW_LOG(NAW_NOTICE,
                        LXM_q ("unknown logging facility: "),
                        LXM_mb(v, vlen));
                break;
            }
            if (!memcmp(facilities->c_name, v, vlen+1)) {
                facility = facilities->c_val;
                break;
            }
            facilities++;
        }
    }

    oldlogptr = naw_log_register(naw_log_syslog);

    naw_logfmt(NAW_INFO, "syslog logging method imported");

    return NAW_OK;
}

static void naw_log_syslog_fini (void)
{
    log_debug (2, "log-syslog: resetting original logging method",0,0,0,0);
    (void) naw_log_register(oldlogptr);
    free(prefix);
}

struct naw_modfuncset naw_set = {
    naw_log_syslog_noop,  /* accept  */
    naw_log_syslog_noop,  /* connect */
    naw_log_syslog_init,  /* init    */
    naw_log_syslog_fini,  /* fini    */
};

