#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/rand.h>
#include <string.h>
#include "misc.h"
#include "naw.h"
#include "naw_log.h"

/* XXX we need:
     - a CRL implementation
*/

struct naw_x509_alloc {
    SSL_CTX *ctx;
    SSL *ssl;
    BIO *bio;
};

static inline int freeret (a, r)
    struct naw_x509_alloc *a;
    int r;
{
    int errcode;

    while ((errcode = ERR_get_error()))
        naw_logfmt(NAW_WARNING, "x509: %s", ERR_error_string(errcode, 0));

    if (a->ssl) { SSL_free(a->ssl); a->bio = 0; }
    if (a->ctx) SSL_CTX_free(a->ctx);
    if (a->bio) BIO_free(a->bio);
    return r;
    /* XXX: deal with error stack, too */
}

static inline int naw_x509_setup(a)
    struct naw_x509_alloc *a;
{
    memset(a, 0, sizeof(struct naw_x509_alloc));

    a->ctx = SSL_CTX_new(TLSv1_method());
    if (!a->ctx) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ctx failed"));
        return freeret(a, NAW_ERR);
    }

    a->ssl = SSL_new(a->ctx);
    if (!a->ssl) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ssl failed"));
        return freeret(a, NAW_ERR);
    }

    a->bio = BIO_new(BIO_s_socket());
    if (!a->bio) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new bio failed"));
        return freeret(a, NAW_ERR);
    }

    return NAW_OK;
}

/***************************************************************************
  naw_test_module_accept

  takes the place of both 'accept' and 'reject' actions.  Only returns YES,
  so 'deny' actions will result in rejection, and 'accept' actions will
  result in access.
 ***************************************************************************/

static int naw_x509_accept (handle, data)
    struct naw_handle *handle;
    struct naw_data *data;
{
    struct naw_x509_alloc a;
    char *v;
    int vlen;
    char *keyfile, *certfile, *cafile;
    int (*ssl_cmd)(SSL *ssl);

    if (handle->type == INIT_SRV) {
        keyfile  = "x509ServerKeyFile";
        certfile = "x509ServerCertFile";
        cafile   = "x509ServerCAFile";
        ssl_cmd  = SSL_accept;
    } else {
        keyfile  = "x509ClientKeyFile";
        certfile = "x509ClientCertFile";
        cafile   = "x509ClientCAFile";
        ssl_cmd  = SSL_connect;
    }
#define common_keyfile  "x509KeyFile"
#define common_certfile "x509CertFile"
#define common_cafile   "x509CAFile"

    memset(&a, 0, sizeof(struct naw_x509_alloc));

    a.ctx = SSL_CTX_new(TLSv1_method());
    if (!a.ctx) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ctx failed"));
        goto ERR_RETURN;
    }

    if (!naw_data_get(data, keyfile, &v, &vlen) ||
        !naw_data_get(data, common_keyfile, &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using key file: "),
                LXM_mb(v,vlen));
        if (!SSL_CTX_use_PrivateKey_file(a.ctx, v, SSL_FILETYPE_PEM)) {
            NAW_LOG(NAW_NOTICE,
                    LXM_q("x509: error reading private key file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }
    }

    if (!naw_data_get(data, certfile, &v, &vlen) ||
        !naw_data_get(data, common_certfile, &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using cert file: "),
                LXM_mb(v,vlen));
        if (!SSL_CTX_use_certificate_file(a.ctx, v, SSL_FILETYPE_PEM)) {
            NAW_LOG(NAW_NOTICE,
                    LXM_q("x509: error reading certificate file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }
    }

    if (!naw_data_get(data, cafile, &v, &vlen) ||
        !naw_data_get(data, common_cafile, &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using ca file: "),
                LXM_mb(v,vlen));
        if (!SSL_CTX_load_verify_locations(a.ctx, v, 0)) {
            NAW_LOG(NAW_NOTICE, LXM_q("x509: error reading CA file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }
    }


    a.ssl = SSL_new(a.ctx);
    if (!a.ssl) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ssl failed"));
        goto ERR_RETURN;
    }

    a.bio = BIO_new(BIO_s_socket());
    if (!a.bio) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new bio failed"));
        goto ERR_RETURN;
    }

    BIO_set_fd(a.bio, handle->sd, BIO_NOCLOSE);

    SSL_set_bio(a.ssl, a.bio, a.bio);
    SSL_set_verify(a.ssl, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, 0);
    if (ssl_cmd(a.ssl) != 1) {
        NAW_LOG(NAW_INFO, LXM_q("x509: ssl negotiation failed"));
        goto ERR_RETURN;
    }
    NAW_LOG(NAW_INFO, LXM_q("x509: ssl negotiation succeeded"));
    //SSL_shutdown(a.ssl);
    return freeret(&a, NAW_YES);

ERR_RETURN:
    return freeret(&a, NAW_ERR);
}

static int naw_x509_init (data)
    struct naw_data *data;
{
    NAW_LOG (NAW_DEBUG2, LXM_q("x509: initializing"));
    SSL_load_error_strings();
    SSL_library_init();
    if (!RAND_status()) {
        NAW_LOG(NAW_NOTICE, LXM_q("x509: not enough random data seeded"));
        /* XXX: seed prng */
    }
    return NAW_OK;
}

static void naw_x509_fini (void)
{
    NAW_LOG (NAW_DEBUG2, LXM_q("x509: unloading"));
    ERR_free_strings();
    /* XXX: RAND_cleanup(); ? */
}

struct naw_modfuncset naw_set = {
    naw_x509_accept,  /* accept  */
    naw_x509_accept,  /* connect */
    naw_x509_init,    /* init    */
    naw_x509_fini,    /* fini    */
};

